#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define __USE_GNU
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>

#define O_NOFOLLOW 0x20000

#ifdef O_NOFOLLOW
#   define O_MODE (O_NOFOLLOW | O_RDWR)
#else
#   define O_MODE (O_RDWR)
#endif


#define IOCTL_READ_MINIMON_DESC             0x1
#define IOCTL_WRITE_MINIMON_IMAGE           0x2
#define IOCTL_KILL_MINIMON                  0x3

#define MAX_ARGC                       15

#define LIN2MINI_WAIT_FOREVER          0xffffffff
#define LIN2MINI_WAIT_X_MS(X)          (X*1000)
#define LIN2MINI_DONT_WAIT             0x0


#define MAP_TA_SEM_VAL_OFFSET          (0x1a98 >> 2)
#define	MAP_IC_SW_INT_OFFSET	       (0xc2d8 >> 2)

#define MINIMON_FREE                   0x0
#define MINIMON_BUSY                   0x1

#define WINMON_CLEAR_VALUE      0xdeadbeef
#define MINIMON_ALIVE_VALUE     0xabc1def2

#define SHELL_MNG_MINIMON_CMD_MINIPING 0x1
#define SHELL_MNG_MINIMON_CMD_MINIRD   0x2
#define SHELL_MNG_MINIMON_CMD_MINIRM   0x3
#define SHELL_MNG_MINIMON_CMD_MINIGO   0x4
#define SHELL_MNG_MINIMON_CMD_MINIKILL 0x10
#define SHELL_MNG_MINIMON_CMD_MINIELD  0x10001      

#define MINIMON_POLL_ADDR_OFFSET       0
#define MINIMON_ARGC_ADDR_OFFSET       0x1
#define MINIMON_ARGV_OFFSET            (0x2+0x10)
#define MINIMON_LAUNCH_OFFSET          0x3d2
#define MINIMON_ERR_OFFSET             0x3d3
#define MINIMON_RESULT_OFFSET          0x3d4
#define MINIMON_INT_PARAM_OFFSET       0x3d5
#define MINIMON_RESERVED2_OFFSET       0x3d6
#define MINIMON_LINUX_PRINTER_OFFSET   0x3d7
#define MINIMON_BUSY_FLAG_OFFSET       0x3d8

#define LOAD_READ_BUFFER_SIZE 256


typedef struct {
      unsigned int dual_core_enabled;
      unsigned int print_buff_addr;
      unsigned int comm_start_addr;
      unsigned int page_mask;
}minimon_mng_descr_t;

typedef union {
 struct {
      unsigned int scratch_number;
      unsigned int size;
      char         data[LOAD_READ_BUFFER_SIZE+1];
 }minimon_image_raw_data;
 struct {
      unsigned int entry;
      unsigned int start;
      unsigned int end;
      unsigned int size;
 }minimon_image_elf_data;
}minimon_mng_image_elf_t;


int handle_cmd(unsigned int cmd, int argc, char ** argv);

unsigned int shell_test_mini_busy(unsigned int *minimon_current_operation);
unsigned int shell_trigger_mini_cmd(  unsigned int cmd, 
                                      unsigned int argc, 
                                      char **argv, 
                                      unsigned int start_addr, 
                                      unsigned int wait_for_result,
                                      unsigned int *minimon_current_operation);
unsigned int shell_get_last_result_from_mini(unsigned int *minimon_result);


int shell_miniping(int argc, char ** argv);
int shell_minird(int argc, char ** argv);
int shell_minirm(int argc, char ** argv);
int shell_minigo(int argc, char ** argv);
int shell_minikill(int argc, char ** argv);
int shell_minield(int argc, char ** argv);


volatile unsigned int *map_chciimt = 0;
volatile unsigned int *minimon_mng_addr = 0;
char                  *MINIMON_ARGV = 0;
unsigned int           fd_mng = 0;

int main(int argc, char **argv)
{
   int                   rc;
   unsigned int          fd_rif;
   volatile unsigned int *vadr_mng = 0;
   minimon_mng_descr_t   minimon_mng_desc; 
   unsigned int          cmd;

   if(argc > 1)
   {
      if(strcmp(argv[1], "miniping") == 0)
      {
         cmd = SHELL_MNG_MINIMON_CMD_MINIPING;
      }
      else if(strcmp(argv[1], "minird") == 0)
      {
         cmd = SHELL_MNG_MINIMON_CMD_MINIRD;
      }
      else if(strcmp(argv[1], "minirm") == 0)
      {
         cmd = SHELL_MNG_MINIMON_CMD_MINIRM;
      }
      else if(strcmp(argv[1], "minield") == 0)
      {
         cmd = SHELL_MNG_MINIMON_CMD_MINIELD;
      }
      else if(strcmp(argv[1], "minigo") == 0)
      {
         cmd = SHELL_MNG_MINIMON_CMD_MINIGO;
      }
      else if(strcmp(argv[1], "minikill") == 0)
      {
         cmd = SHELL_MNG_MINIMON_CMD_MINIKILL;
      }
      else
      {
         printf("minimon management: unknown command %s.\n",argv[1]);
         printf("minimon management: Supported commands - miniping, minird, minirm, minield, minigo and minikill\n");
         exit(1);
      }
   }
   else
   {
         printf("minimon management: no command passed \n");
         printf("minimon management: Supported commands - miniping, minird, minirm, minield, minigo and minikill\n");
         exit(1);
   }
   
   if ((fd_mng = open("/dev/wp_mini2", O_MODE))<0 )
   {
      printf("open failed");
      return -1;
   }

   if ((fd_rif = open("/dev/wp_mini3", O_MODE))<0 )
   {
      printf("open failed");
      return -1;
   }

   if (ioctl(fd_mng, IOCTL_READ_MINIMON_DESC, &minimon_mng_desc) < 0)
   {
      printf("ioctl failed");
      return -1;
   }   

   vadr_mng = mmap(0, 0x2000,
               PROT_WRITE, MAP_SHARED,
               fd_mng, 0);

   if (vadr_mng == MAP_FAILED)
   {
      printf("mmap failed");
      return -1;
   }   
 
   minimon_mng_addr     = ((unsigned int*)((unsigned int)vadr_mng     + (minimon_mng_desc.comm_start_addr & (~minimon_mng_desc.page_mask))));

   map_chciimt = mmap(0, 0x00040000,
               PROT_WRITE, MAP_SHARED,
               fd_rif, 0);

   if (map_chciimt == MAP_FAILED)
   {
      printf("mmap failed");
      return -1;
   }   

   rc = handle_cmd(cmd, (argc-1), &(argv[1]));
   
   if(rc != 0)
      printf("minimon management: operation failed\n");
   
   return rc;
}


int handle_cmd(unsigned int cmd, int argc, char ** argv)
{
   int rc = -1;
   
   switch(cmd)
   {
      case SHELL_MNG_MINIMON_CMD_MINIPING:
         rc = shell_miniping(argc, argv);
         break;
      case SHELL_MNG_MINIMON_CMD_MINIRD:
         rc = shell_minird(argc, argv);
         break;
      case SHELL_MNG_MINIMON_CMD_MINIRM:
         rc = shell_minirm(argc, argv);
         break;
      case SHELL_MNG_MINIMON_CMD_MINIGO:
         rc = shell_minigo(argc, argv);
         break;
      case SHELL_MNG_MINIMON_CMD_MINIKILL:
         rc = shell_minikill(argc, argv);
         break;
      case SHELL_MNG_MINIMON_CMD_MINIELD:
         rc = shell_minield(argc, argv);
         break;
      default:
         printf("minimon management: unknown command %d\n",cmd);
   }
   
   return rc;
}

int shell_miniping(int argc, char ** argv)
{
   unsigned int minimon_current_operation = 0;
   unsigned int minimon_result; 
   
   unsigned int rc = shell_trigger_mini_cmd(SHELL_MNG_MINIMON_CMD_MINIPING, 0, 0, 0, 
                                            LIN2MINI_WAIT_X_MS(500), 
                                            &minimon_current_operation);
   if(rc != 0)
   {
      printf("minimon management: Minimon is busy. last command sent was 0x%08x\n", minimon_current_operation);
      return -1;      
   }
   else
   {
      rc = shell_get_last_result_from_mini(&minimon_result);      
      if(rc == 0)
      {
         if(minimon_result == MINIMON_ALIVE_VALUE)
         {
            printf("Minimon is ready...\n");
            return 0;            
         }
         return (-1);
      }
      printf("Minimon reported unknown status 0x%08x\n",rc);      
      return (-1);
   }      
   return 0;   

}

int shell_minird(int argc, char ** argv)
{
   unsigned int minimon_current_operation;   
   unsigned int minimon_result; 
   unsigned int rc = shell_trigger_mini_cmd(SHELL_MNG_MINIMON_CMD_MINIRD, 
                                              argc, argv, 0, 
                                              LIN2MINI_WAIT_X_MS(500),
                                              &minimon_current_operation);

   if(rc != 0)
   {
      printf("minimon management: Minimon is busy. last command sent was 0x%08x\n", minimon_current_operation);
      return -1;      
   }
   else
   {      
      rc = shell_get_last_result_from_mini(&minimon_result);      
      //printf( "minimon_result = %08x\n", minimon_result );
      if(rc != 0)
         printf("Minimon reported error\n");      
      return rc;
   }      
   
   return 0;   

}

int shell_minirm(int argc, char ** argv)
{
   unsigned int minimon_current_operation;   
   unsigned int rc = shell_trigger_mini_cmd(SHELL_MNG_MINIMON_CMD_MINIRM, argc, argv, 0, 
                                      LIN2MINI_DONT_WAIT, &minimon_current_operation);
   if(rc != 0)
   {      
      printf("minimon management: Minimon is busy. last command sent was 0x%08x\n", minimon_current_operation);
      return -1;      
   }
   
   return 0;   

}

int shell_minigo(int argc, char ** argv)
{
   unsigned int minimon_current_operation;
   unsigned int rc;

   if(argc != 2)
   {
      printf("minimon management: no start address to launch application\n");
      return -1;
   }

   rc = shell_trigger_mini_cmd(SHELL_MNG_MINIMON_CMD_MINIGO, argc, argv, 0, 
                                 LIN2MINI_DONT_WAIT, &minimon_current_operation);
   if(rc != 0)
   {      
      printf("minimon management: Minimon is busy. last command sent was 0x%08x\n", minimon_current_operation);
      return -1;      
   }

   return 0;   
}

int shell_minikill(int argc, char ** argv)
{
//   SetMiniCmd( MINIMON_INT_KILL_CMD );
//   map_chciimt[MAP_IC_SW_INT_OFFSET] = 0x4;//trigger SW2 interrupt
   int rc;
   
   minimon_mng_addr[MINIMON_INT_PARAM_OFFSET]=SHELL_MNG_MINIMON_CMD_MINIKILL;
   __asm__ __volatile__("sync");
   
   if ((rc = ioctl(fd_mng, IOCTL_KILL_MINIMON, NULL)) < 0)
   {
      printf("ioctl failed: IOCTL_KILL_MINIMON\n");
      return rc;
   }   
   return 0;   
}

#define WEB_PACKETRAM_BASE (0x30000000 + 0x100)
int shell_minield(int argc, char ** argv)
{
   FILE *srec_file = 0;
   minimon_mng_image_elf_t image;
   int rc;

   image.minimon_image_raw_data.scratch_number = WEB_PACKETRAM_BASE;

   if(argc != 2)
   {
      printf("minimon management: no file name passed for minield\n");
      return -1;
   }

   srec_file = fopen(argv[1], "r");
   
   if(! srec_file )
   {
      printf("minimon management: failed to open file %s\n", argv[1]);
      return -1;
   }
   
   do
   {
      memset(image.minimon_image_raw_data.data, 0, sizeof(image.minimon_image_raw_data.data));
      image.minimon_image_raw_data.size = fread(image.minimon_image_raw_data.data, sizeof(char), LOAD_READ_BUFFER_SIZE, srec_file);
      //even if read 0 bytes - send notification to the kernel about EOF
      if ((rc = ioctl(fd_mng, IOCTL_WRITE_MINIMON_IMAGE, &image)) < 0)
      {
         if(rc != ENODATA)
         {
            printf("ioctl failed: IOCTL_LOAD_IMAGE\n");
            fclose(srec_file);
            return -1;
         }
      }   
   }while(image.minimon_image_raw_data.size == LOAD_READ_BUFFER_SIZE);
   
   fclose(srec_file);

   if(rc == 0)
   {
      printf("ELF Read: copying 0x%08x bytes to address 0x%08x-0x%08x\n",
             image.minimon_image_elf_data.size, image.minimon_image_elf_data.start , image.minimon_image_elf_data.end); 
      printf("Entry point: 0x%08x\n\n",image.minimon_image_elf_data.entry);
      return rc;
   }
   else
   {
      printf("Opreation failed\n");
      return -1;
   }
   
   return 0;
}


unsigned int shell_test_mini_busy(unsigned int *minimon_current_operation)
{
#if 0
   unsigned int hw_sem_test = map_chciimt[MAP_TA_SEM_VAL_OFFSET];
   
   if(hw_sem_test & 0x00000020)
   {
      *minimon_current_operation = minimon_mng_addr[MINIMON_POLL_ADDR_OFFSET]; 
      return MINIMON_BUSY;
   }
   else 
      return MINIMON_FREE;
#endif

   unsigned int rc = minimon_mng_addr[MINIMON_BUSY_FLAG_OFFSET];
   if(rc == MINIMON_BUSY)
   {
      *minimon_current_operation = minimon_mng_addr[MINIMON_POLL_ADDR_OFFSET]; 
      return MINIMON_BUSY;
   }
   else 
      return MINIMON_FREE;
}

unsigned int shell_trigger_mini_cmd(  unsigned int cmd, 
                                      unsigned int argc, 
                                      char **argv, 
                                      unsigned int start_addr, 
                                      unsigned int wait_for_result,
                                      unsigned int *minimon_current_operation)
{
   unsigned int rc;
   int i, j;
   
   rc = shell_test_mini_busy(minimon_current_operation);   

   if(rc == MINIMON_FREE)
   {
      //copy the data to the noncached com pointers:
      MINIMON_ARGV = (char*)&(minimon_mng_addr[MINIMON_ARGV_OFFSET]);
      
      for( i = 0, j = 0; i < argc; i++, j += 256 )
         strcpy( (MINIMON_ARGV + j) ,argv[i] );

      minimon_mng_addr[MINIMON_ARGC_ADDR_OFFSET]   = (volatile unsigned long)argc;
      minimon_mng_addr[MINIMON_LAUNCH_OFFSET]      = (volatile unsigned long)start_addr;
      minimon_mng_addr[MINIMON_ERR_OFFSET]         = (volatile unsigned long)WINMON_CLEAR_VALUE; /*clear Minimon's last action return code*/
      minimon_mng_addr[MINIMON_RESULT_OFFSET]      = (volatile unsigned long)WINMON_CLEAR_VALUE; /*clear Minimon's last action result*/
      minimon_mng_addr[MINIMON_POLL_ADDR_OFFSET]   = (volatile unsigned long)cmd;
      
      if(wait_for_result == LIN2MINI_DONT_WAIT)
         return rc;      
      else if(wait_for_result == LIN2MINI_WAIT_FOREVER)
         while((volatile unsigned long) minimon_mng_addr[MINIMON_ERR_OFFSET] == WINMON_CLEAR_VALUE);
      else
         usleep(wait_for_result);
   }
   return rc;   

}

unsigned int shell_get_last_result_from_mini(unsigned int *minimon_result)
{
   *minimon_result = minimon_mng_addr[MINIMON_RESULT_OFFSET];
//   printf( "*minimon_result = %08x\n",*minimon_result);
   return minimon_mng_addr[MINIMON_ERR_OFFSET];
}
