#include <stdio.h>
#define __USE_GNU
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>

//#define O_NOFOLLOW 0x20000 /* we want to get signals here */

#ifdef O_NOFOLLOW
#   define O_MODE (O_NOFOLLOW | O_RDWR)
#else
#   define O_MODE (O_RDWR)
#endif

#define SIGNAL_DUAL_PRINT 64

#define IOCTL_READ_MINIMON_DESC             1

#define MINIMON_PRINT_REQUEST   0xd0d0d000
#define MINIMON_PRINT_COMPLETED 0xcafecaff
#define MINIMON_PRINT_REQUEST_MASK 0xffffff00

#define MINIMON_POLL_ADDR_OFFSET         0
#define MINIMON_ARGC_ADDR_OFFSET         1
#define MINIMON_ARGV_OFFSET              2
#define MINIMON_LAUNCH_OFFSET        0x3d2
#define MINIMON_ERR_OFFSET           0x3d3
#define MINIMON_RESULT_OFFSET        0x3d4
#define MINIMON_INT_PARAM_OFFSET     0x3d5
#define MINIMON_RESERVED2_OFFSET     0x3d6
#define MINIMON_LINUX_PRINTER_OFFSET 0x3d7

#define MAX_BUFFER_LEN 256
#define NUM_OF_TC     3
#define NUM_OF_CORES  2

#define CORE_TC_2_IDX( CORE, TC ) (((CORE)*(NUM_OF_TC))+(TC))
#define MINIMON_PRINT_GET_CORE(x) ((x>>4) & 0x0000000f)
#define MINIMON_PRINT_GET_TC(x)   (x & 0x0000000f)

void minimon_printer_sig_handler(int signum, siginfo_t* info);

typedef struct TCPrintBuffer
{
	char buffer[MAX_BUFFER_LEN];
	int  n_chars;
} TCPrintBuffer;

typedef struct {
      unsigned int dual_core_enabled;
      unsigned int print_buff_addr;
      unsigned int comm_start_addr;
      unsigned int page_mask;
}minimon_mng_descr_t;

static unsigned int *vadr_mng = 0;
static volatile unsigned int *unaligned_vadr_mng = 0;
static unsigned int *vadr_printer = 0;
static volatile unsigned int *unaligned_vadr_printer = 0;
static unsigned int minimon_printer_message = MINIMON_PRINT_COMPLETED;
static unsigned int core, tc, index, print_buff_addr;

int main(int argc, char **argv)
{
   int  fd_printer;
   int  fd_mng;
   minimon_mng_descr_t     minimon_mng_desc;
   struct sigaction siginfo;

   
   if ((fd_printer = open("/dev/wp_mini1", O_RDWR))<0 )
   {
      printf("minimon printer: open failed\n");
      return -1;
   }

   if ((fd_mng = open("/dev/wp_mini2", O_MODE))<0 )
   {
      printf("minimon printer: open failed\n");
      return -1;
   }

   if (ioctl(fd_mng, IOCTL_READ_MINIMON_DESC, &minimon_mng_desc) < 0)
   {
      printf("minimon printer: ioctl failed\n");
      return -1;
   }   

   if(minimon_mng_desc.dual_core_enabled == 0)
   {
      printf("minimon printer: Dual core disabled\n");
      return 0;
   }

   vadr_mng = mmap(0, 0x2000,
               PROT_WRITE, MAP_SHARED,
               fd_mng, 0);

   if (vadr_mng == MAP_FAILED)
   {
      printf("minimon printer: mmap failed\n");
      return -1;
   }   

   vadr_printer = mmap(0, 0x2000,
               PROT_WRITE, MAP_SHARED,
               fd_printer, 0);

   if (vadr_printer == MAP_FAILED)
   {
      printf("minimon printer: mmap failed\n");
      return -1;
   }

   sigaction(SIGNAL_DUAL_PRINT, NULL, &siginfo);
   siginfo.sa_sigaction = (void *)minimon_printer_sig_handler;
   sigemptyset(&siginfo.sa_mask);
   siginfo.sa_flags = SA_SIGINFO;

   if (sigaction(SIGNAL_DUAL_PRINT, &siginfo, NULL) == -1) {
      printf(("Minimon Printer Error: failed to regsiter signal handler\n"));
      return -1;
   }

 
   unaligned_vadr_printer = ((unsigned int*)((unsigned int)vadr_printer + (minimon_mng_desc.print_buff_addr & (~minimon_mng_desc.page_mask))));
   unaligned_vadr_mng     = ((unsigned int*)((unsigned int)vadr_mng     + (minimon_mng_desc.comm_start_addr & (~minimon_mng_desc.page_mask))));

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#if 0
   printf("vadr_mng at 0x%08x value 0x%08x\n", (unsigned int)vadr_mng, *vadr_mng);
   printf("vadr_printer at 0x%08x value 0x%08x\n", (unsigned int)vadr_printer, *vadr_printer);

   printf("unaligned_vadr_mng at 0x%08x value 0x%08x\n", (unsigned int)unaligned_vadr_mng, *unaligned_vadr_mng);
   printf("unaligned_vadr_printer at 0x%08x value 0x%08x\n", (unsigned int)unaligned_vadr_printer, *unaligned_vadr_printer);


   printf("printer vaddr from kernel 0x%08x\n",(unsigned int)vadr_printer);
   printf("Minimon print buffers are at 0x%08x\n", (unsigned int)unaligned_vadr_printer);
   printf("Minimon mng   buffers are at 0x%08x\n", (unsigned int)unaligned_vadr_mng);
   printf("minimon desc:\n");
   printf("dual_core_enabled: 0x%08x\n",minimon_mng_desc.dual_core_enabled);
   printf("print_buff_addr  : 0x%08x\n",minimon_mng_desc.print_buff_addr);
   printf("page_mask        : 0x%08x\n",minimon_mng_desc.page_mask);
   printf("comm_start_addr  : 0x%08x\n",minimon_mng_desc.comm_start_addr);
#endif
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        
   while(1)
   {
      sleep(1000);
   }
   
   printf("\n\n********************************\n");
   printf("  MINIMON PRINTER - EXIT NOW !!!\n");
   printf("********************************\n\n");
   
   return 0;
}

void minimon_printer_sig_handler(int signum, siginfo_t* info)
{
   minimon_printer_message = unaligned_vadr_mng[MINIMON_LINUX_PRINTER_OFFSET];      
   if((minimon_printer_message & MINIMON_PRINT_REQUEST_MASK) != MINIMON_PRINT_REQUEST)
   {
      printf("minimon_printer: Invalid signal. No MINIMON_PRINT_REQUEST token read\n");
      return;
   }      

   core = MINIMON_PRINT_GET_CORE(minimon_printer_message);
   tc   = MINIMON_PRINT_GET_TC(minimon_printer_message);
   index = CORE_TC_2_IDX(core, tc);
   print_buff_addr = (unsigned int)unaligned_vadr_printer + (index * sizeof(TCPrintBuffer));      
   printf("%s", (char*)print_buff_addr);
   fflush(stdout);
   unaligned_vadr_mng[MINIMON_LINUX_PRINTER_OFFSET] = MINIMON_PRINT_COMPLETED;
   __asm__ __volatile__ ("sync");      
}
